import torch

from ..layers import ConfidenceWeights, NeighborhoodAggregation
from ..utils import common_functions as c_f
from .base import BaseHook
from .features import FeaturesAndLogitsHook


class ATDOCHook(BaseHook):
    """
    Creates pseudo labels for the target domain
    using k-nearest neighbors. Then computes a
    classification loss based on these pseudo labels.

    Implementation of
    [Domain Adaptation with Auxiliary Target Domain-Oriented Classifier](https://arxiv.org/abs/2007.04171).
    """

    def __init__(
        self, dataset_size, feature_dim, num_classes, k=5, loss_fn=None, **kwargs
    ):
        """
        Arguments:
            dataset_size: The number of samples in the target dataset.
            feature_dim: The feature dimensionality, i.e at each iteration
                the features should be size ```(N, D)``` where N is batch size and
                D is ```feature_dim```.
            num_classes: The number of class labels in the target dataset.
            k: The number of nearest neighbors used to determine each
                sample's pseudolabel
            loss_fn: The classification loss function.
                If ```None``` it defaults to
                ```torch.nn.CrossEntropyLoss```.
        """
        super().__init__(**kwargs)
        self.labeler = NeighborhoodAggregation(
            dataset_size, feature_dim, num_classes, k=k
        )
        self.weighter = ConfidenceWeights()
        self.loss_fn = c_f.default(
            loss_fn, torch.nn.CrossEntropyLoss, {"reduction": "none"}
        )
        self.hook = FeaturesAndLogitsHook(domains=["target"])

    def call(self, losses, inputs):
        """"""
        outputs = self.hook(losses, inputs)[1]
        [features, logits] = c_f.extract(
            [outputs, inputs],
            c_f.filter(self.hook.out_keys, "", ["_features$", "_logits$"]),
        )
        pseudo_labels, neighbor_logits = self.labeler(
            features, logits, update=True, idx=inputs["target_sample_idx"]
        )
        loss = self.loss_fn(logits, pseudo_labels)
        weights = self.weighter(neighbor_logits)
        loss = torch.mean(weights * loss)
        return {"pseudo_label_loss": loss}, outputs

    def _loss_keys(self):
        """"""
        return ["pseudo_label_loss"]

    def _out_keys(self):
        """"""
        return self.hook.out_keys



from ..containers import KeyEnforcer
from ..hooks import AlignerPlusCHook, RTNHook
from ..utils.common_functions import check_domain
from .base_adapter import BaseGCAdapter


class Aligner(BaseGCAdapter):
    """
    Wraps [AlignerPlusCHook][pytorch_adapt.hooks.aligners.AlignerPlusCHook].
    """

    hook_cls = AlignerPlusCHook

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(opts=list(self.optimizers.values()), **hook_kwargs)


class RTN(Aligner):
    """
    Wraps [RTNHook][pytorch_adapt.hooks.rtn].
    """

    hook_cls = RTNHook

    def inference_default(self, x, domain=None):
        domain = check_domain(self, domain)
        features, logits = super().inference_default(x, domain)
        if domain == 0:
            return features, self.models["residual_model"](logits)
        return features, logits

    def get_key_enforcer(self):
        ke = super().get_key_enforcer()
        ke.requirements["models"].append("residual_model")
        ke.requirements["optimizers"].append("residual_model")
        ke.requirements["misc"] = ["feature_combiner"]
        return ke



from typing import List

from ..layers import GradientReversal
from ..utils import common_functions as c_f
from .base import BaseHook, BaseWrapperHook
from .cdan import CDANDomainHook, get_cdan_features_hooks, get_entropy_reducer
from .classification import CLossHook, SoftmaxHook
from .domain import DomainLossHook, FeaturesForDomainLossHook
from .features import (
    CombinedFeaturesHook,
    FeaturesAndLogitsHook,
    FeaturesHook,
    LogitsHook,
)
from .optimizer import OptimizerHook, SummaryHook
from .utils import (
    ApplyFnHook,
    AssertHook,
    BaseWrapperHook,
    ChainHook,
    OnlyNewOutputsHook,
)


class GradientReversalHook(ApplyFnHook):
    def __init__(self, weight=1, **kwargs):
        super().__init__(fn=GradientReversal(weight), **kwargs)


class GradientReversalLocallyHook(BaseWrapperHook):
    def __init__(self, apply_to: List[str], *hooks: BaseHook, weight=1, **kwargs):
        super().__init__(**kwargs)
        grl = GradientReversalHook(weight=weight, apply_to=apply_to)
        self.hook = OnlyNewOutputsHook(ChainHook(grl, *hooks, overwrite=True))


class SoftmaxGradientReversalHook(BaseWrapperHook):
    def __init__(self, apply_to, weight=1, **kwargs):
        super().__init__(**kwargs)
        self.hook = ChainHook(
            SoftmaxHook(apply_to=apply_to),
            GradientReversalHook(weight=weight, apply_to=apply_to),
            overwrite=True,
        )


class SoftmaxGradientReversalLocallyHook(BaseWrapperHook):
    def __init__(self, apply_to, *hooks, weight=1, **kwargs):
        super().__init__(**kwargs)
        sgrl = SoftmaxGradientReversalHook(apply_to, weight)
        self.hook = OnlyNewOutputsHook(ChainHook(sgrl, *hooks, overwrite=True))


class DANNHook(BaseWrapperHook):
    """
    Implementation of
    [Domain-Adversarial Training of Neural Networks](https://arxiv.org/abs/1505.07818).

    This includes the model optimization step.
    """

    def __init__(
        self,
        opts,
        weighter=None,
        reducer=None,
        pre=None,
        pre_d=None,
        post_d=None,
        pre_g=None,
        post_g=None,
        gradient_reversal=None,
        gradient_reversal_weight=1,
        use_logits=False,
        f_hook=None,
        d_hook=None,
        c_hook=None,
        domain_loss_hook=None,
        d_hook_allowed="_dlogits$",
        **kwargs
    ):
        """
        Arguments:
            opts: List of optimizers for updating the models.
            weighter: Weights the losses before backpropagation.
                If ```None``` then it defaults to
                [```MeanWeighter```][pytorch_adapt.weighters.mean_weighter.MeanWeighter]
            reducer: Reduces loss tensors.
                If ```None``` then it defaults to
                [```MeanReducer```][pytorch_adapt.hooks.reducers.MeanReducer]
            pre: List of hooks that will be executed at the very
                beginning of each iteration.
            pre_d: List of hooks that will be executed after
                gradient reversal, but before the domain loss.
            post_d: List of hooks that will be executed after
                gradient reversal, and after the domain loss.
            pre_g: List of hooks that will be executed outside of
                the gradient reversal step, and before the generator
                and classifier loss.
            post_g: List of hooks that will be executed after
                the generator and classifier losses.
            gradient_reversal: Called before all D hooks, including
                ```pre_d```.
            use_logits: If ```True```, then D receives the output of C
                instead of the output of G.
            f_hook: The hook used for computing features and logits.
                If ```None``` then it defaults to
                [```FeaturesForDomainLossHook```][pytorch_adapt.hooks.domain.FeaturesForDomainLossHook]
            d_hook: The hook used for computing discriminator logits.
                If ```None``` then it defaults to
                [```DLogitsHook```][pytorch_adapt.hooks.features.DLogitsHook]
            c_hook: The hook used for computing the classifiers's loss.
                If ```None``` then it defaults to
                [```CLossHook```][pytorch_adapt.hooks.classification.CLossHook]
            domain_loss_hook: The hook used for computing the domain loss.
                If ```None``` then it defaults to
                [```DomainLossHook```][pytorch_adapt.hooks.domain.DomainLossHook].
            d_hook_allowed: A regex string that specifies the allowed
                output names of the discriminator block.
        """
        super().__init__(**kwargs)
        [pre, pre_d, post_d, pre_g, post_g] = c_f.many_default(
            [pre, pre_d, post_d, pre_g, post_g], [[], [], [], [], []]
        )
        f_hook = c_f.default(
            f_hook, FeaturesForDomainLossHook, {"use_logits": use_logits}
        )
        gradient_reversal = c_f.default(
            gradient_reversal,
            GradientReversalHook,
            {"weight": gradient_reversal_weight, "apply_to": f_hook.out_keys},
        )
        c_hook = c_f.default(c_hook, CLossHook, {})
        domain_loss_hook = c_f.default(
            domain_loss_hook, DomainLossHook, {"f_hook": f_hook, "d_hook": d_hook}
        )

        disc_hook = AssertHook(
            OnlyNewOutputsHook(
                ChainHook(
                    gradient_reversal,
                    *pre_d,
                    domain_loss_hook,
                    *post_d,
                    overwrite=[1],
                )
            ),
            d_hook_allowed,
        )
        gen_hook = ChainHook(*pre_g, c_hook, *post_g)

        hook = ChainHook(*pre, f_hook, disc_hook, gen_hook)
        hook = OptimizerHook(hook, opts, weighter, reducer)
        s_hook = SummaryHook({"total_loss": hook})
        self.hook = ChainHook(hook, s_hook)


class DANNLogitsHook(DANNHook):
    def __init__(self, **kwargs):
        f_hook = FeaturesForDomainLossHook(use_logits=True)
        super().__init__(f_hook=f_hook, **kwargs)


class DANNSoftmaxLogitsHook(DANNHook):
    def __init__(self, **kwargs):
        f_hook = FeaturesForDomainLossHook(use_logits=True)
        gradient_reversal = SoftmaxGradientReversalHook(apply_to=f_hook.out_keys)
        super().__init__(f_hook=f_hook, gradient_reversal=gradient_reversal, **kwargs)


class GradientReversalThenEntropyReducer(BaseWrapperHook):
    def __init__(
        self, detach_entropy_reducer=True, gradient_reversal_weight=1, **kwargs
    ):
        super().__init__(**kwargs)
        reducer = get_entropy_reducer(
            apply_to=["src_domain_loss", "target_domain_loss"],
            detach_weights=detach_entropy_reducer,
        )
        apply_to = ["src_imgs_features_logits", "target_imgs_features_logits"]
        reducer = GradientReversalLocallyHook(
            apply_to, reducer, weight=gradient_reversal_weight
        )
        self.hook = ChainHook(LogitsHook(), reducer)


class DANNEHook(DANNHook):
    def __init__(
        self, detach_entropy_reducer=True, gradient_reversal_weight=1, **kwargs
    ):
        reducer = GradientReversalThenEntropyReducer(
            detach_entropy_reducer, gradient_reversal_weight
        )
        super().__init__(
            reducer=reducer, gradient_reversal_weight=gradient_reversal_weight, **kwargs
        )


class CDANNEHook(DANNEHook):
    def __init__(self, softmax=True, **kwargs):
        f_hook, fc_hook, _ = get_cdan_features_hooks(softmax)
        f_hook = ChainHook(f_hook, fc_hook)
        domain_loss_hook = CDANDomainHook(
            loss_prefix="", detach_features=False, reverse_labels=False, softmax=True
        )
        super().__init__(f_hook=f_hook, domain_loss_hook=domain_loss_hook, **kwargs)



import os

import torch
from torchvision import datasets as torch_datasets

from .base_dataset import BaseDataset
from .utils import check_img_paths, check_length


class Office31Full(BaseDataset):
    """
    A small dataset consisting of 31 classes in 3 domains:
    amazon, dslr, webcam.
    """

    def __init__(self, root: str, domain: str, transform):
        """
        Arguments:
            root: The dataset must be located at ```<root>/office31```
            domain: One of the 3 domains
            transform: The image transform applied to each sample.
        """

        super().__init__(domain=domain)
        self.transform = transform
        self.dataset = torch_datasets.ImageFolder(
            os.path.join(root, "office31", domain, "images"), transform=self.transform
        )
        check_length(self, {"amazon": 2817, "dslr": 498, "webcam": 795}[domain])

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        return self.dataset[idx]


class Office31(BaseDataset):
    """
    A custom train/test split of Office31Full.
    """

    def __init__(self, root: str, domain: str, train: bool, transform, **kwargs):
        """
        Arguments:
            root: The dataset must be located at ```<root>/office31```
            domain: One of the 3 domains
            train: Whether or not to use the training set.
            transform: The image transform applied to each sample.
        """

        super().__init__(domain=domain, **kwargs)
        if not isinstance(train, bool):
            raise TypeError("train should be True or False")
        name = "train" if train else "test"
        labels_file = os.path.join(root, "office31", f"{domain}_{name}.txt")
        img_dir = os.path.join(root, "office31")

        with open(labels_file) as f:
            content = [line.rstrip().split(" ") for line in f]
        self.img_paths = [os.path.join(img_dir, x[0]) for x in content]
        check_img_paths(img_dir, self.img_paths, domain)
        check_length(
            self,
            {
                "amazon": {"train": 2253, "test": 564}[name],
                "dslr": {"train": 398, "test": 100}[name],
                "webcam": {"train": 636, "test": 159}[name],
            }[domain],
        )
        self.labels = [int(x[1]) for x in content]
        self.transform = transform



from typing import Callable

import torch
import torch.nn.functional as F

from ..utils import common_functions as c_f


def mcd_loss(out1, out2, dist_fn):
    return dist_fn(F.softmax(out1, dim=1), F.softmax(out2, dim=1))


class MCDLoss(torch.nn.Module):
    """
    Implementation of the loss function used in
    [Maximum Classifier Discrepancy for Unsupervised Domain Adaptation](https://arxiv.org/abs/1712.02560).
    """

    def __init__(self, dist_fn: Callable[[torch.Tensor], torch.Tensor] = None):
        """
        Arguments:
            dist_fn: Computes the mean distance between two softmaxed tensors.
                If ```None```, then ```torch.nn.L1Loss``` is used.
        """
        super().__init__()
        self.dist_fn = c_f.default(dist_fn, torch.nn.L1Loss, {})

    def forward(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        """
        Arguments:
            x: a batch of class logits
            y: the other batch of class logits
        Returns:
            The discrepancy between the two batches of class logits.
        """
        return mcd_loss(x, y, self.dist_fn)


def general_mcd_loss(*x, p=1):
    x = [torch.nn.functional.softmax(i, dim=1) for i in x]
    loss = 0
    for i in range(len(x)):
        for j in range(i + 1, len(x)):
            loss += F(x[i], x[j])
    return loss


class GeneralMCDLoss(torch.nn.Module):
    def __init__(self, p=1):
        super().__init__()
        self.p = p

    def forward(self, *x):
        """"""
        return general_mcd_loss(*x, p=self.p)



from ..containers import KeyEnforcer
from ..hooks import SymNetsHook
from ..utils.common_functions import check_domain
from .base_adapter import BaseGCAdapter


class SymNets(BaseGCAdapter):
    """
    Wraps [SymNetsHook][pytorch_adapt.hooks.symnets].
    """

    hook_cls = SymNetsHook

    def inference_default(self, x, domain):
        domain = check_domain(self, domain)
        features = self.models["G"](x)
        logits = self.models["C"](features)[domain]
        return features, logits

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(
            g_opts=[self.optimizers["G"]], c_opts=[self.optimizers["C"]], **hook_kwargs
        )



import torch
import torch.nn.functional as F

from .concat_softmax import ConcatSoftmax
from .utils import split_half


class SymNetsCategoryLoss(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.softmax_fn = ConcatSoftmax()

    # x and y are the first and second halves of "p^st"
    def forward(self, x, y, src_labels):
        x = self.softmax_fn(x, y)
        x, y = split_half(x, dim=1)
        x_loss = F.cross_entropy(x, src_labels)
        y_loss = F.cross_entropy(y, src_labels)
        return x_loss + y_loss


class SymNetsCategoryLossListInput(SymNetsCategoryLoss):
    def forward(self, x, src_labels):
        return super().forward(*x, src_labels)



from typing import Dict, List

import torch

from ..utils import common_functions as c_f
from ..weighters import BaseWeighter, MeanWeighter
from .base import BaseHook, BaseWrapperHook
from .reducers import BaseReducer, MeanReducer


class OptimizerHook(BaseHook):
    """
    1. Executes the wrapped hook
    2. Zeros all gradients
    3. Backpropagates the loss
    4. Steps the optimizer
    """

    def __init__(
        self,
        hook: BaseHook,
        optimizers: List[torch.optim.Optimizer],
        weighter: BaseWeighter = None,
        reducer: BaseReducer = None,
        **kwargs
    ):
        """
        Arguments:
            hook: the hook that computes the losses
            optimizers: a list of optimizers that will be used
                to update model weights
            weighter: weights the returned losses and outputs a
                single value on which ```.backward()``` is called.
                If ```None```, then it defaults to
                [```MeanWeighter```][pytorch_adapt.weighters.mean_weighter.MeanWeighter].
            reducer: a hook that reduces any unreduced losses to a single value.
                If ```None```, then it defaults to
                [```MeanReducer```][pytorch_adapt.hooks.reducers.MeanReducer].
        """
        super().__init__(**kwargs)
        self.hook = hook
        self.optimizers = optimizers
        self.weighter = c_f.default(weighter, MeanWeighter, {})
        self.reducer = c_f.default(reducer, MeanReducer, {})
        self.loss_components = {}

    def call(self, losses, inputs):
        """"""
        losses, outputs = self.hook(losses, inputs)
        c_f.assert_dicts_are_disjoint(inputs, outputs)
        losses, new_outputs = self.reducer(losses, {**inputs, **outputs})
        outputs.update(new_outputs)
        loss, self.loss_components = self.weighter(losses)
        c_f.zero_back_step(loss, self.optimizers)
        return {}, outputs

    def _loss_keys(self):
        """"""
        return []

    def _out_keys(self):
        """"""
        return c_f.join_lists([self.hook.out_keys, self.reducer.out_keys])

    def extra_repr(self):
        return c_f.extra_repr(self, ["weighter"])


class SummaryHook(BaseHook):
    """
    Repackages losses into a dictionary format useful for logging.
    This should be used only at the very end of each
    iteration, i.e. it should be the last sub-hook
    in a [ChainHook][pytorch_adapt.hooks.utils.ChainHook].
    """

    def __init__(self, optimizers: Dict[str, OptimizerHook], **kwargs):
        """
        Arguments:
            optimizers: A dictionary of optimizer hooks.
                The losses computed inside these hooks
                will be packaged into nested dictionaries.
        """
        super().__init__(**kwargs)
        self.optimizers = optimizers

    def call(self, losses, inputs):
        """"""
        losses = {}
        for k, v in self.optimizers.items():
            losses[k] = v.loss_components
        return losses, {}

    def _loss_keys(self):
        """"""
        return list(self.optimizers.keys())

    def _out_keys(self):
        """"""
        return []



import math

import torch
import torch.nn.functional as F
from pytorch_metric_learning.distances import CosineSimilarity

from ..layers import EntropyLoss
from .base_validator import BaseValidator


# https://arxiv.org/pdf/2108.10860.pdf
class SNDValidator(BaseValidator):
    """
    Implementation of
    [Tune it the Right Way: Unsupervised Validation of Domain Adaptation via Soft Neighborhood Density](https://arxiv.org/abs/2108.10860)
    """

    def __init__(self, layer="preds", T=0.05, batch_size=1000, **kwargs):
        super().__init__(**kwargs)
        self.layer = layer
        self.T = T
        self.dist_fn = CosineSimilarity()
        self.entropy_fn = EntropyLoss(after_softmax=True, return_mean=False)
        self.batch_size = batch_size

    def compute_score(self, target_train):
        features = target_train[self.layer]
        n = features.shape[0]
        all_entropies = []
        for s in range(0, n, self.batch_size):
            e = s + self.batch_size
            L = features[s:e]
            sim_mat = self.dist_fn(L, features)
            sim_mat = self.mask_out_self(sim_mat, s)
            sim_mat = F.softmax(sim_mat / self.T, dim=1)
            all_entropies.append(self.entropy_fn(sim_mat))
        all_entropies = torch.cat(all_entropies, dim=0)
        if len(all_entropies) != len(features):
            raise ValueError("all_entropies should have same length as input features")
        return torch.mean(all_entropies).item()

    def mask_out_self(self, sim_mat, start_idx):
        num_rows, num_cols = sim_mat.shape
        mask = torch.ones(num_rows, num_cols, dtype=torch.bool)
        rows = torch.arange(num_rows)
        cols = rows + start_idx
        mask[rows, cols] = False
        return sim_mat[mask].view(num_rows, num_cols - 1)



import os

import torch
from torchvision import datasets as torch_datasets

from .base_dataset import BaseDataset
from .utils import check_img_paths, check_length


class OfficeHomeFull(BaseDataset):
    def __init__(self, root: str, domain: str, transform):
        super().__init__(domain=domain)
        self.transform = transform
        self.dataset = torch_datasets.ImageFolder(
            os.path.join(root, "officehome", domain),
            transform=self.transform,
        )
        check_length(
            self, {"art": 2427, "clipart": 4365, "product": 4439, "real": 4357}[domain]
        )

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        return self.dataset[idx]


class OfficeHome(BaseDataset):
    def __init__(self, root: str, domain: str, train: bool, transform, **kwargs):
        super().__init__(domain=domain, **kwargs)
        if not isinstance(train, bool):
            raise TypeError("train should be True or False")
        name = "train" if train else "test"
        labels_file = os.path.join(root, "officehome", f"{domain}_{name}.txt")
        img_dir = os.path.join(root, "officehome")

        with open(labels_file) as f:
            content = [line.rstrip().split(" ") for line in f]
        self.img_paths = [os.path.join(img_dir, x[0]) for x in content]
        check_img_paths(img_dir, self.img_paths, domain)
        check_length(
            self,
            {
                "art": {"train": 1941, "test": 486}[name],
                "clipart": {"train": 3492, "test": 873}[name],
                "product": {"train": 3551, "test": 888}[name],
                "real": {"train": 3485, "test": 872}[name],
            }[domain],
        )
        self.labels = [int(x[1]) for x in content]
        self.transform = transform



from ..layers import MMDLoss
from ..utils import common_functions as c_f
from .aligners import AlignerHook
from .base import BaseWrapperHook
from .classification import CLossHook
from .features import BaseFeaturesHook, CombinedFeaturesHook, FeaturesAndLogitsHook
from .losses import TargetEntropyHook
from .optimizer import OptimizerHook, SummaryHook
from .utils import ChainHook


class ResidualHook(BaseFeaturesHook):
    def __init__(
        self,
        in_suffixes=None,
        out_suffixes=None,
        **kwargs,
    ):
        in_suffixes = c_f.default(in_suffixes, ["_imgs_features_logits"])
        out_suffixes = c_f.default(out_suffixes, ["_plus_residual"])
        super().__init__(
            model_name="residual_model",
            in_suffixes=in_suffixes,
            out_suffixes=out_suffixes,
            **kwargs,
        )


def rtn_aligner_key_map(fc_hook):
    strs = c_f.filter(fc_hook.out_keys, "", ["^src", "^target"])
    return {k: v for k, v in zip(strs, ["src_imgs_features", "target_imgs_features"])}


class RTNAlignerHook(BaseWrapperHook):
    def __init__(self, loss_fn=None, **kwargs):
        super().__init__(**kwargs)
        loss_fn = c_f.default(loss_fn, MMDLoss, {})
        fl_hook = FeaturesAndLogitsHook()
        fc_hook = CombinedFeaturesHook()
        key_map = rtn_aligner_key_map(fc_hook)
        aligner_hook = AlignerHook(loss_fn=loss_fn, layer="features", key_map=key_map)
        self.hook = ChainHook(fl_hook, fc_hook, aligner_hook)


class RTNLogitsHook(BaseWrapperHook):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        fl_hook = FeaturesAndLogitsHook()
        entropy_hook = TargetEntropyHook()
        residual_hook = ResidualHook(domains=["src"])
        closs_hook = CLossHook(
            key_map={residual_hook.out_keys[0]: "src_imgs_features_logits"}
        )
        self.hook = ChainHook(fl_hook, entropy_hook, residual_hook, closs_hook)


class RTNHook(BaseWrapperHook):
    """
    Implementation of
    [Unsupervised Domain Adaptation with Residual Transfer Networks](https://arxiv.org/abs/1602.04433).
    """

    def __init__(
        self,
        opts,
        weighter=None,
        reducer=None,
        pre=None,
        post=None,
        aligner_loss_fn=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        [pre, post] = c_f.many_default([pre, post], [[], []])
        hook = ChainHook(*pre, RTNAlignerHook(aligner_loss_fn), RTNLogitsHook(), *post)
        hook = OptimizerHook(hook, opts, weighter, reducer)
        s_hook = SummaryHook({"total_loss": hook})
        self.hook = ChainHook(hook, s_hook)



from collections import defaultdict

import torch
from ignite.engine import Engine
from ignite.metrics import Metric

# These decorators helps with distributed settings
from ignite.metrics.metric import reinit__is_reduced, sync_all_reduce


class DictionaryAccumulator(Metric):
    def __init__(self):
        super().__init__()
        self.reset()

    @reinit__is_reduced
    def reset(self):
        super().reset()
        self.accumulated = defaultdict(list)

    @reinit__is_reduced
    def update(self, output):
        for k, v in output.items():
            self.accumulated[k].append(v)

    @sync_all_reduce("accumulated")
    def compute(self):
        for k, v in self.accumulated.items():
            self.accumulated[k] = torch.cat(v, dim=0)
        return self.accumulated

    @torch.no_grad()
    def iteration_completed(self, engine: Engine) -> None:
        self.update(engine.state.output)



import os
from abc import ABC, abstractmethod

import torch
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..frameworks import Ignite
from ..validators.multiple_validators import MultipleValidators
from . import common_functions as c_f


class BaseSaver(ABC):
    def __init__(self, folder):
        self.folder = folder

    @abstractmethod
    def save(self):
        pass

    @abstractmethod
    def load(self):
        pass


class ContainerSaver(BaseSaver):
    def save(self, container, epoch, best_epoch, prefix=""):
        suffixes = [str(epoch)]
        if epoch == best_epoch:
            suffixes.append("best")
        for suffix in suffixes:
            for k, v in container.items():
                if not c_f.has_state_dict(v):
                    continue
                filename = c_f.class_as_prefix(
                    container, k, prefix=prefix, suffix=suffix
                )
                c_f.save_torch_module(v, self.folder, f"{filename}.pth")

    def load(self, container, suffix, prefix=""):
        for k, v in container.items():
            if not c_f.has_state_dict(v):
                continue
            if suffix == "latest":
                basename = c_f.class_as_prefix(container, k, prefix=prefix)
                suffix, _ = pml_cf.latest_version(
                    self.folder, string_to_glob=f"{basename}*.pth"
                )
            filename = c_f.class_as_prefix(container, k, prefix=prefix, suffix=suffix)
            c_f.load_torch_module(v, self.folder, f"{filename}.pth")
        return suffix

    def delete(self, container, keep, prefix=""):
        to_keep = [(prefix, keep), (prefix, "best")]
        for k, v in container.items():
            if not c_f.has_state_dict(v):
                continue
            basename = c_f.class_as_prefix(container, k)
            c_f.delete_all_but(self.folder, basename, ".pth", to_keep=to_keep)

    def copy(self, container, from_suffix, to_suffix, prefix=""):
        c_f.LOGGER.info(f"Copying {from_suffix} to {to_suffix}")
        for k, v in container.items():
            if not c_f.has_state_dict(v):
                continue
            filenames = [
                c_f.class_as_prefix(container, k, prefix=prefix, suffix=x)
                for x in [from_suffix, to_suffix]
            ]
            filenames = [os.path.join(self.folder, f"{x}.pth") for x in filenames]
            c_f.copy_file(filenames[0], filenames[1])


class MultipleContainersSaver(BaseSaver):
    def __init__(self, container_saver=None, **kwargs):
        super().__init__(**kwargs)
        self.container_saver = c_f.default(container_saver, ContainerSaver, kwargs)

    def save(self, container, epoch, best_epoch, prefix=""):
        # save each child container
        for v in container.values():
            self.container_saver.save(v, epoch, best_epoch, prefix)

    def load(self, container, suffix, prefix=""):
        # load each child container
        for v in container.values():
            self.container_saver.load(v, suffix, prefix)

    def delete(self, container, keep, prefix=""):
        for v in container.values():
            self.container_saver.delete(v, keep, prefix)

    def copy(self, container, from_suffix, to_suffix, prefix=""):
        for v in container.values():
            self.container_saver.copy(v, from_suffix, to_suffix, prefix)


class AdapterSaver(BaseSaver):
    def __init__(self, container_saver=None, keep_every_version=False, **kwargs):
        super().__init__(**kwargs)
        self.container_saver = c_f.default(
            container_saver, MultipleContainersSaver, kwargs
        )
        self.keep_every_version = keep_every_version

    def save(self, adapter, epoch, best_epoch, prefix=""):
        self.container_saver.save(adapter.containers, epoch, best_epoch, prefix=prefix)
        if not self.keep_every_version:
            self.container_saver.delete(adapter.containers, keep=epoch, prefix=prefix)
        if self.keep_every_version and epoch != best_epoch:
            self.container_saver.copy(
                adapter.containers, best_epoch, "best", prefix=prefix
            )

    def load(self, adapter, suffix, prefix=""):
        self.container_saver.load(adapter.containers, suffix, prefix=prefix)


class ValidatorSaver(BaseSaver):
    def save(self, validator, prefix=""):
        if isinstance(validator, MultipleValidators):
            self.save_multiple(validator, prefix)
        else:
            self.save_single(validator, prefix)

    def load(self, validator, prefix=""):
        if isinstance(validator, MultipleValidators):
            self.load_multiple(validator, prefix)
        else:
            self.load_single(validator, prefix)

    def save_single(self, validator, prefix=""):
        # cast best_epoch from int64 to regular int so it can be json-parsed
        best_epoch = validator.best_epoch
        if best_epoch is not None:
            best_epoch = int(best_epoch)
        output = {
            "best_score": validator.best_score,
            "best_epoch": best_epoch,
        }
        filename = c_f.class_as_prefix(validator, "best", prefix=prefix)
        c_f.save_json(output, self.folder, f"{filename}.json")
        for name in self.saveable_attrs:
            filename = c_f.class_as_prefix(validator, name, prefix=prefix)
            c_f.save_npy(getattr(validator, name), self.folder, f"{filename}.npy")

    def load_single(self, validator, prefix=""):
        for name in self.saveable_attrs:
            filename = c_f.class_as_prefix(validator, name, prefix=prefix)
            setattr(validator, name, c_f.load_npy(self.folder, f"{filename}.npy"))

    def save_multiple(self, validator, prefix=""):
        for k, v in validator.validators.items():
            child_prefix = c_f.class_as_prefix(validator, k, prefix=prefix)
            self.save_single(v, child_prefix)
        self.save_single(validator, prefix)

    def load_multiple(self, validator, prefix=""):
        for k, v in validator.validators.items():
            child_prefix = c_f.class_as_prefix(validator, k, prefix=prefix)
            self.load_single(v, child_prefix)
        self.load_single(validator, prefix)

    @property
    def saveable_attrs(self):
        return ["raw_score_history", "score_history", "epochs"]


class IgniteSaver(BaseSaver):
    def filename(self):
        return c_f.full_path(self.folder, "ignite_engine.pth")

    def save(self, engine):
        torch.save(engine.state_dict(), self.filename())

    def load(self, engine):
        engine.load_state_dict(torch.load(self.filename()))


class Saver:
    def __init__(
        self,
        adapter_saver=None,
        validator_saver=None,
        stat_getter_saver=None,
        ignite_saver=None,
        **kwargs,
    ):
        self.adapter_saver = c_f.default(adapter_saver, AdapterSaver, kwargs)
        self.validator_saver = c_f.default(validator_saver, ValidatorSaver, kwargs)
        self.stat_getter_saver = c_f.default(stat_getter_saver, ValidatorSaver, kwargs)
        self.ignite_saver = c_f.default(ignite_saver, IgniteSaver, kwargs)

    def save_adapter(self, adapter, epoch, best_epoch):
        c_f.LOGGER.info(f"Saving adapter to {self.adapter_saver.folder}")
        self.adapter_saver.save(adapter, epoch, best_epoch, "adapter")

    def load_adapter(self, adapter, suffix):
        c_f.LOGGER.info(f"Loading adapter from {self.adapter_saver.folder}")
        self.adapter_saver.load(adapter, suffix, "adapter")

    def save_validator(self, validator):
        c_f.LOGGER.info(f"Saving validator to {self.validator_saver.folder}")
        self.validator_saver.save(validator, "validator")

    def load_validator(self, validator):
        c_f.LOGGER.info(f"Loading validator from {self.validator_saver.folder}")
        self.validator_saver.load(validator, "validator")

    def save_stat_getter(self, stat_getter):
        c_f.LOGGER.info(f"Saving stat_getter to {self.stat_getter_saver.folder}")
        self.stat_getter_saver.save(stat_getter, "stat_getter")

    def load_stat_getter(self, stat_getter):
        c_f.LOGGER.info(f"Loading stat_getter from {self.stat_getter_saver.folder}")
        self.stat_getter_saver.load(stat_getter, "stat_getter")

    def save_ignite(self, engine):
        c_f.LOGGER.info(f"Saving Ignite engine to {self.ignite_saver.folder}")
        self.ignite_saver.save(engine)

    def load_ignite(self, engine):
        c_f.LOGGER.info(f"Loading Ignite engine from {self.ignite_saver.folder}")
        self.ignite_saver.load(engine)

    def load_all(
        self,
        adapter=None,
        validator=None,
        stat_getter=None,
        framework=None,
        suffix="latest",
    ):
        if adapter:
            self.load_adapter(adapter, suffix)
        if validator:
            self.load_validator(validator)
        if stat_getter:
            self.load_stat_getter(stat_getter)
        if framework:
            if isinstance(framework, Ignite):
                self.load_ignite(framework.trainer)



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import torch

from ..utils import common_functions as c_f


class BaseHook(ABC):
    """All hooks extend ```BaseHook```"""

    def __init__(
        self,
        loss_prefix: str = "",
        loss_suffix: str = "",
        out_prefix: str = "",
        out_suffix: str = "",
        key_map: Dict[str, str] = None,
    ):
        """
        Arguments:
            loss_prefix: prepended to all new loss keys
            loss_suffix: appended to all new loss keys
            out_prefix: prepended to all new output keys
            out_suffix: appended to all new output keys
            key_map: a mapping from ```input_key``` to ```new_key```.
                For example, if key_map = {"A": "B"}, and the input dict to ```__call__``` is {"A": 5},
                then the input will be converted to {"B": 5} before being consumed. Before exiting ```__call__```,
                the mapping is undone so the input context is preserved.
                In other words, {"B": 5} will be converted back to {"A": 5}.
        """
        if any(
            not isinstance(x, str)
            for x in [loss_prefix, loss_suffix, out_prefix, out_suffix]
        ):
            raise TypeError("loss prefix/suffix and out prefix/suffix must be strings")
        self.loss_prefix = loss_prefix
        self.loss_suffix = loss_suffix
        self.out_prefix = out_prefix
        self.out_suffix = out_suffix
        self.key_map = c_f.default(key_map, {})
        self.in_keys = []

    def __call__(self, losses, inputs):
        try:
            inputs = c_f.map_keys(inputs, self.key_map)
            x = self.call(losses, inputs)
            if isinstance(x, (bool, np.bool_)):
                return x
            elif isinstance(x, tuple):
                losses, outputs = x
                outputs = replace_mapped_keys(outputs, self.key_map)
                inputs = replace_mapped_keys(inputs, self.key_map)
                losses = wrap_keys(losses, self.loss_prefix, self.loss_suffix)
                outputs = wrap_keys(outputs, self.out_prefix, self.out_suffix)
                self.check_losses_and_outputs(losses, outputs, inputs)
                return losses, outputs
            else:
                raise TypeError(
                    f"Output is of type {type(x)}, but should be bool or tuple"
                )
        except Exception as e:
            if not isinstance(e, KeyError):
                c_f.append_error_message(e, self.str_for_error_msg(n=1))
            raise

    @abstractmethod
    def call(
        self, losses: Dict[str, Any], inputs: Dict[str, Any]
    ) -> Union[Tuple[Dict[str, Any], Dict[str, Any]], bool]:
        """
        This must be implemented by the child class
        Arguments:
            losses: previously computed losses
            inputs: holds everything else: tensors, models etc.
        Returns:
            Either a tuple of (losses, outputs) that will be merged with the input context,
            or a boolean
        """
        pass

    @abstractmethod
    def _loss_keys(self) -> List[str]:
        """
        This must be implemented by the child class
        Returns:
            The names of the losses that will be added to the context.
        """
        pass

    @property
    def loss_keys(self):
        return list(
            set(wrap_keys(self._loss_keys(), self.loss_prefix, self.loss_suffix))
        )

    @abstractmethod
    def _out_keys(self) -> List[str]:
        """
        This must be implemented by the child class
        Returns:
            The names of the outputs that will be added to the context.
        """
        pass

    @property
    def out_keys(self):
        x = replace_mapped_keys(self._out_keys(), self.key_map)
        return list(set(wrap_keys(x, self.out_prefix, self.out_suffix)))

    def set_in_keys(self, in_keys):
        self.in_keys = in_keys

    def __repr__(self):
        return c_f.nice_repr(self, self.extra_repr(), self.children_repr())

    def extra_repr(self):
        return ""

    def children_repr(self):
        all_hooks = c_f.attrs_of_type(self, BaseHook)
        all_modules = c_f.attrs_of_type(self, torch.nn.Module)
        c_f.assert_dicts_are_disjoint(all_hooks, all_modules)
        return {**all_hooks, **all_modules}

    def str_for_error_msg(self, x=None, n=None):
        e = str(self if x is None else x)
        if n is not None:
            e = "\n".join(e.split("\n")[:n])
            e += "\n...\n"
        return f"\nERROR occuring in:\n{e}"

    def check_losses_and_outputs(self, losses, outputs, inputs):
        check_keys_are_present(self, self.loss_keys, [losses], "loss_keys", "losses")
        check_keys_are_present(
            self, self.out_keys, [inputs, outputs], "out_keys", "inputs or outputs"
        )
        check_keys_are_present(self, losses, self.loss_keys, "loss_keys", "losses")
        check_keys_are_present(self, outputs, self.out_keys, "outputs", "out_keys")


def wrap_keys(x, prefix, suffix):
    if prefix == "" and suffix == "":
        return x
    if isinstance(x, dict):
        new_x = {}
        for k, v in x.items():
            new_x[f"{prefix}{k}{suffix}"] = v
        return new_x
    elif c_f.is_list_or_tuple(x):
        return [f"{prefix}{y}{suffix}" for y in x]
    else:
        raise TypeError("input to wrap_keys should be dict, list, or tuple")


def replace_mapped_keys(outputs, key_map):
    if len(key_map) == 0:
        return outputs
    reverse_key_map = {v: k for k, v in key_map.items()}
    if isinstance(outputs, dict):
        return {
            c_f.map_keys_substrings(k, reverse_key_map): v for k, v in outputs.items()
        }
    elif c_f.is_list_or_tuple(outputs):
        return [c_f.map_keys_substrings(k, reverse_key_map) for k in outputs]


def check_keys_are_present(cls, x, y, x_name, y_name):
    set_x = c_f.to_set(x)
    set_y = c_f.to_set(y)
    diff = set_x - set_y
    if len(diff) > 0:
        msg = f"keys {diff} are in {x_name} but are not present in {y_name} of {c_f.cls_name(cls)}"
        raise KeyError(msg)


class BaseConditionHook(BaseHook):
    """The base class for hooks that return a boolean"""

    def _loss_keys(self):
        """"""
        return []

    def _out_keys(self):
        """"""
        return []


class BaseWrapperHook(BaseHook):
    """A simple wrapper for calling ```self.hook```,
    which should be defined in the child's ```__init__``` function."""

    def call(self, *args, **kwargs):
        """"""
        return self.hook(*args, **kwargs)

    def _loss_keys(self):
        """"""
        return self.hook.loss_keys

    def _out_keys(self):
        """"""
        return self.hook.out_keys



import torch
from pytorch_metric_learning.utils import common_functions as pml_cf
from pytorch_metric_learning.utils import stat_utils

from ..layers import SilhouetteScore
from .base_validator import BaseValidator


class SilhouetteScoreValidator(BaseValidator):
    def __init__(self, layer="features", **kwargs):
        super().__init__(**kwargs)
        self.layer = layer
        self.silhouette_fn = SilhouetteScore()

    def compute_score(self, target_train):
        return get_silhouette_score(
            target_train[self.layer],
            target_train["logits"],
            silhouette_fn=self.silhouette_fn,
        )


### https://github.com/lr94/abas/blob/master/model_selection.py ###
def get_centroids(data, labels, num_classes):
    centroids = torch.zeros((num_classes, data.shape[1]), device=data.device)
    for cid in range(num_classes):
        matches = labels == cid
        if torch.any(matches):
            centroids[cid] = torch.mean(data[matches], dim=0)

    return centroids


def get_silhouette_score(feats, logits, silhouette_fn):
    plabels = torch.argmax(logits, dim=1)
    num_classes = logits.shape[1]
    centroids = get_centroids(feats, plabels, num_classes)
    clabels = get_kmeans(feats, num_classes, centroids)
    return silhouette_fn(feats, clabels)


def get_kmeans(x, num_classes, init_centroids):
    import faiss

    device = x.device
    x = pml_cf.to_numpy(x)
    init_centroids = pml_cf.to_numpy(init_centroids)
    km = faiss.Kmeans(x.shape[1], num_classes, gpu=True)
    km.train(x, init_centroids=init_centroids)
    _, I = km.index.search(x, 1)
    I = torch.from_numpy(I).squeeze(1)
    return pml_cf.to_device(I, device=device)



import inspect
import json
import os
from abc import ABC, abstractmethod
from typing import Callable, Dict, List

import numpy as np
import torch
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f


class BaseValidator(ABC):
    """
    The parent class of all validators.

    The main purpose of validators is to give an estimate
    of target domain accuracy, without having access to
    class labels.
    """

    def __init__(
        self,
        normalizer: Callable[[np.ndarray], np.ndarray] = None,
        key_map: Dict[str, str] = None,
        ignore_epoch: int = 0,
    ):
        """
        Arguments:
            normalizer: A function that receives the current unnormalized
                score history, and returns a normalized version of the
                score history. If ```None```, then it defaults to
                no normalization.
            key_map: A mapping from ```<new_split_names>``` to
                ```<original_split_names>```. For example,
                [```AccuracyValidator```][pytorch_adapt.validators.AccuracyValidator]
                expects ```src_val``` by default. When used with one of the
                [```frameworks```](../frameworks/index.md), this default
                indicates that data related to the ```src_val``` split should be retrieved.
                If you instead want to compute accuracy for the ```src_train``` split,
                you would set the ```key_map``` to ```{"src_train": "src_val"}```.
            ignore_epoch: This epoch will ignored when determining
                the best scoring epoch. The default of 0 is meant to be
                reserved for the initial model (before training has begun).
        """
        self.normalizer = c_f.default(normalizer, return_raw)
        self.score_history = np.array([])
        self.raw_score_history = np.array([])
        self.epochs = np.array([], dtype=int)
        self.key_map = c_f.default(key_map, {})
        self.ignore_epoch = ignore_epoch
        pml_cf.add_to_recordable_attributes(
            self,
            list_of_names=["latest_score", "best_score", "latest_epoch", "best_epoch"],
        )

    def _required_data(self):
        args = inspect.getfullargspec(self.compute_score).args
        args.remove("self")
        return args

    @property
    def required_data(self) -> List[str]:
        """
        Returns:
            A list of dataset split names.
        """
        output = set(self._required_data()) - set(self.key_map.values())
        output = list(output)
        for k, v in self.key_map.items():
            output.append(k)
        return output

    @abstractmethod
    def compute_score(self):
        pass

    def score(self, epoch: int, **kwargs: Dict[str, torch.Tensor]) -> float:
        """
        Arguments:
            epoch: The epoch to be scored.
            **kwargs: A mapping from dataset split name to
                dictionaries containing:

                - ```"features"```
                - ```"logits"```
                - ```"preds"```
                - ```"domain"```
                - ```"src_labels"``` (if available)
        """
        kwargs = self.kwargs_check(epoch, kwargs)
        score = self.compute_score(**kwargs)
        self.append_to_history_and_normalize(score, epoch)
        return self.latest_score

    def kwargs_check(self, epoch, kwargs):
        if epoch in self.epochs:
            raise ValueError(f"Epoch {epoch} has already been evaluated")
        if kwargs.keys() != set(self.required_data):
            raise ValueError(
                f"Input to compute_score has keys = {kwargs.keys()} but should have keys {self.required_data}"
            )
        return c_f.map_keys(kwargs, self.key_map)

    def append_to_history_and_normalize(self, score, epoch):
        self.raw_score_history = np.append(self.raw_score_history, score)
        self.epochs = np.append(self.epochs, epoch)
        self.score_history = self.normalizer(self.raw_score_history)

    @property
    def score_history_ignore_epoch(self):
        return remove_ignore_epoch(self.score_history, self.epochs, self.ignore_epoch)

    @property
    def epochs_ignore_epoch(self):
        return remove_ignore_epoch(self.epochs, self.epochs, self.ignore_epoch)

    def has_valid_history(self, ignore_ignore_epoch=True):
        x = (
            self.score_history_ignore_epoch
            if ignore_ignore_epoch
            else self.score_history
        )
        return len(x) > 0 and np.isfinite(x).any()

    @property
    def best_score(self) -> float:
        """
        Returns:
            The best score, ignoring ```self.ignore_epoch```.
            Returns ```None``` if no valid scores are available.
        """
        if self.has_valid_history():
            return self.score_history_ignore_epoch[self.best_idx]

    @property
    def best_epoch(self) -> int:
        """
        Returns:
            The best epoch, ignoring ```self.ignore_epoch```.
            Returns ```None``` if no valid epochs are available.
        """
        if self.has_valid_history():
            return self.epochs_ignore_epoch[self.best_idx]

    @property
    def best_idx(self) -> int:
        """
        Returns:
            The index of the best score in ```self.score_history_ignore_epoch```.
            Returns ```None``` if no valid epochs are available.
        """
        if self.has_valid_history():
            return (
                np.nanargmax(self.score_history_ignore_epoch)
                if self.maximize
                else np.nanargmin(self.score_history_ignore_epoch)
            )

    @property
    def latest_epoch(self) -> int:
        """
        Returns:
            The latest epoch, including ```self.ignore_epoch```.
            Returns ```None``` if no epochs have been scored.
        """
        if self.has_valid_history(False):
            return self.epochs[-1]

    @property
    def latest_score(self) -> float:
        """
        Returns:
            The latest score, including ```self.ignore_epoch```.
            Returns ```None``` if no epochs have been scored.
        """
        if self.has_valid_history(False):
            return self.score_history[-1]

    @property
    def latest_is_best(self) -> bool:
        """
        Returns:
            ```False``` if the latest epoch was not the best
            scoring epoch, or if the latest epoch is ```ignore_epoch```.
            ```True``` otherwise.
        """
        if self.has_valid_history(False):
            return self.best_epoch == self.latest_epoch
        return False

    @property
    def maximize(self) -> bool:
        """
        Returns:
            ```True``` if a higher validation score indicates a better model.
            The default is ```True```.
        """
        return True

    def __repr__(self):
        return c_f.nice_repr(self, self.extra_repr(), {})

    def extra_repr(self):
        return c_f.extra_repr(
            self, ["latest_score", "best_score", "best_epoch", "required_data"]
        )


def score_dict_to_str(d):
    return "\n\t" + "\n\t".join([f"{k}: {v}" for k, v in d.items()])


def remove_ignore_epoch(x, epochs, ignore_epoch):
    if ignore_epoch is not None:
        return x[epochs != ignore_epoch]
    return x


def return_raw(raw_score_history):
    return raw_score_history



import itertools

import numpy as np
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f
from .base_validator import BaseValidator


class MultipleValidators(BaseValidator):
    def __init__(self, validators, weights=None, **kwargs):
        super().__init__(**kwargs)
        self.validators = c_f.enumerate_to_dict(validators)
        self.weights = c_f.default(
            weights, {k: 1 if v.maximize else -1 for k, v in self.validators.items()}
        )
        self.weights = c_f.enumerate_to_dict(self.weights)
        pml_cf.add_to_recordable_attributes(
            self, list_of_names=["validators", "weights"]
        )

    def _required_data(self):
        output = [v.required_data for v in self.validators.values()]
        output = list(itertools.chain(*output))
        return list(set(output))

    def compute_score(self):
        pass

    def score(self, epoch, **kwargs):
        kwargs = self.kwargs_check(epoch, kwargs)
        for v in self.validators.values():
            v.score(epoch, **c_f.filter_kwargs(kwargs, v.required_data))
        self.append_to_history_and_normalize(epoch)
        return self.latest_score

    def append_to_history_and_normalize(self, epoch):
        self.raw_score_history = []
        for k, v in self.validators.items():
            self.raw_score_history.append(v.score_history * self.weights[k])
        self.raw_score_history = sum(self.raw_score_history)
        self.epochs = np.append(self.epochs, epoch)
        self.score_history = self.normalizer(self.raw_score_history)

    def __repr__(self):
        return c_f.nice_repr(self, self.extra_repr(), self.validators)

    def extra_repr(self):
        x = super().extra_repr()
        x += f"\n{c_f.extra_repr(self, ['weights'])}"
        return x



from typing import Any, Dict

import numpy as np
import torch

from ..utils import common_functions as c_f
from .source_dataset import SourceDataset
from .target_dataset import TargetDataset


class CombinedSourceAndTargetDataset(torch.utils.data.Dataset):
    """
    Wraps a source dataset and a target dataset.
    """

    def __init__(self, source_dataset: SourceDataset, target_dataset: TargetDataset):
        """
        Arguments:
            source_dataset:
            target_dataset:
        """

        self.source_dataset = source_dataset
        self.target_dataset = target_dataset

    def __len__(self) -> int:
        """
        Returns:
            The length of the target dataset.
        """
        return len(self.target_dataset)

    def __getitem__(self, idx) -> Dict[str, Any]:
        """
        Arguments:
            idx: The index of the target dataset. The source index is picked randomly.
        Returns:
            A dictionary containing both source and target data.
            The source keys start with "src", and the target keys start with "target".
        """
        target_data = self.target_dataset[idx]
        src_data = self.source_dataset[self.get_random_src_idx()]
        c_f.assert_dicts_are_disjoint(src_data, target_data)
        return {**src_data, **target_data}

    def get_random_src_idx(self):
        return np.random.choice(len(self.source_dataset))

    def __repr__(self):
        return c_f.nice_repr(
            self,
            "",
            {
                "source_dataset": self.source_dataset,
                "target_dataset": self.target_dataset,
            },
        )



import torch

from ..utils import common_functions as c_f


# reference: https://github.com/wgchang/DSBN/blob/master/model/dsbn.py
class AdaptiveBatchNorm(torch.nn.Module):
    def __init__(
        self, num_features=None, num_domains=2, affine_domain=None, bns=None, **kwargs
    ):
        super().__init__()
        if affine_domain is not None and not isinstance(affine_domain, int):
            raise TypeError("affine_domain must either be None or an int")
        if bns is None:
            bns = []
            for i in range(num_domains):
                bns.append(self.get_batchnorm_class()(num_features, **kwargs))
        self.bns = torch.nn.ModuleList(bns)
        self.affine_domain = affine_domain
        self.curr_domain = None
        self.set_affine_params(delete=True)

    def reset_running_stats(self):
        for bn in self.bns:
            bn.reset_running_stats()

    def reset_parameters(self):
        for bn in self.bns:
            bn.reset_parameters()

    def set_curr_domain(self, domain):
        self.curr_domain = domain

    def set_affine_params(self, delete=False):
        if self.affine_domain is not None:
            source_bn = self.bns[self.affine_domain]
            weight = source_bn.weight.data
            bias = source_bn.bias.data
            for i in range(len(self.bns)):
                if i != self.affine_domain:
                    target_bn = self.bns[i]
                    if delete:
                        del target_bn.weight
                        del target_bn.bias
                    target_bn.weight = weight
                    target_bn.bias = bias

    def forward(self, x, domain=None):
        domain = c_f.default(domain, self.curr_domain)
        if domain is None:
            raise TypeError(
                "If input domain is None, then curr_domain must be set via set_curr_domain"
            )
        self.set_affine_params(delete=False)
        output = self.bns[domain](x)
        self.set_curr_domain(None)
        return output


class AdaptiveBatchNorm2d(AdaptiveBatchNorm):
    def get_batchnorm_class(self):
        return torch.nn.BatchNorm2d


# https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
# https://stackoverflow.com/a/56407442
def update(existingAggregate, newValues):
    newValues = newValues.permute(0, 2, 3, 1).flatten(start_dim=0, end_dim=2)
    (count, mean, M2) = existingAggregate
    count += newValues.shape[0]
    # newvalues - oldMean
    delta = newValues - mean
    mean += torch.sum(delta / count, dim=0)
    # newvalues - newMean
    delta2 = newValues - mean
    M2 += torch.sum(delta * delta2, dim=0)
    return (count, mean, M2)


def finalize(existingAggregate):
    (count, mean, M2) = existingAggregate
    (mean, variance, sampleVariance) = (mean, M2 / count, M2 / (count - 1))
    if count < 2:
        raise ValueError("Can't finalize when count is less than 2")
    else:
        return (mean, variance, sampleVariance)


# original https://github.com/ptrblck/pytorch_misc/blob/master/batch_norm_manual.py
class PopulationBatchNorm2d(torch.nn.BatchNorm2d):
    def __init__(
        self, num_features, eps=1e-5, affine=True, init_mean=None, init_var=None
    ):
        super().__init__(
            num_features=num_features,
            eps=eps,
            momentum=None,
            affine=affine,
            track_running_stats=True,
        )
        self.total_n = 0
        self.register_buffer("M2", torch.zeros(num_features))
        self.register_buffer(
            "final_mean", c_f.default(init_mean, torch.zeros(num_features))
        )
        self.register_buffer(
            "final_var", c_f.default(init_var, torch.ones(num_features))
        )

    def finalize(self):
        try:
            self.final_mean, self.final_var, _ = finalize(
                (self.total_n, self.running_mean, self.M2)
            )
        except ValueError:
            pass

    def reset_running_stats(self):
        super().reset_running_stats()
        self.total_n = 0
        for attr in ["M2", "final_mean", "final_var"]:
            if hasattr(self, attr):
                if attr == "final_var":
                    getattr(self, attr).fill_(1)
                else:
                    getattr(self, attr).zero_()

    def forward(self, input):
        self._check_input_dim(input)
        # calculate running estimates
        if self.training:
            with torch.no_grad():
                self.total_n, self.running_mean, self.M2 = update(
                    (self.total_n, self.running_mean, self.M2), input
                )

        mean = self.final_mean
        var = self.final_var

        input = (input - mean[None, :, None, None]) / (
            torch.sqrt(var[None, :, None, None] + self.eps)
        )
        if self.affine:
            input = (
                input * self.weight[None, :, None, None]
                + self.bias[None, :, None, None]
            )

        return input


def convert_bn_to_adabn(model, affine_domain, bn_type):
    changes = []
    for child_name, child in model.named_children():
        if isinstance(child, (torch.nn.BatchNorm1d, torch.nn.BatchNorm3d)):
            raise TypeError(
                f"child is type {type(child)}, which is not currently supported for conversion to adabn."
            )
        if isinstance(child, torch.nn.BatchNorm2d):
            num_features = child.num_features
            if bn_type == PopulationBatchNorm2d:
                other_bn = bn_type(
                    num_features,
                    init_mean=child.running_mean.clone(),
                    init_var=child.running_var.clone(),
                )
            elif bn_type == torch.nn.BatchNorm2d:
                other_bn = bn_type(num_features)
            new_bn = AdaptiveBatchNorm2d(
                affine_domain=affine_domain, bns=[child, other_bn]
            )
            setattr(model, child_name, new_bn)
        convert_bn_to_adabn(child, affine_domain, bn_type)


def collect_all_bn(model, bn_type):
    x = []
    for m in model.children():
        if isinstance(m, bn_type):
            x.append(m)
        x.extend(collect_all_bn(m, bn_type))
    return x


def collect_all_bn_if_not_done(model, bn_type):
    list_name = f"list_of_{bn_type.__name__}"
    if not hasattr(model, list_name):
        x = collect_all_bn(model, bn_type)
        setattr(model, list_name, x)
    return list_name


def set_curr_domain(model, domain, bn_type):
    list_name = collect_all_bn_if_not_done(model, bn_type)
    for v in getattr(model, list_name):
        v.set_curr_domain(domain)


def finalize_bn(model, bn_type):
    list_name = collect_all_bn_if_not_done(model, bn_type)
    for v in getattr(model, list_name):
        v.finalize()


def set_bn_layer_to_train(model, layer_num, bn_type):
    list_name = collect_all_bn_if_not_done(model, bn_type)
    list_of_bn = getattr(model, list_name)
    set_layer = None
    for i, v in enumerate(list_of_bn):
        if i == layer_num:
            v.train()
            set_layer = i
        else:
            v.eval()
    c_f.LOGGER.info(f"There are {len(list_of_bn)} bn layers")
    if set_layer is not None:
        c_f.LOGGER.info(f"Set bn layer {set_layer} to train mode")
    else:
        c_f.LOGGER.info(f"Did not set any bn layers to train mode.")



import torch


def covariance(x):
    batch_size = x.shape[0]
    mm1 = torch.mm(x.t(), x)
    cols_summed = torch.sum(x, dim=0)
    mm2 = torch.mm(cols_summed.unsqueeze(1), cols_summed.unsqueeze(0))
    return (1.0 / (batch_size - 1)) * (mm1 - (1.0 / batch_size) * mm2)


class CORALLoss(torch.nn.Module):
    """
    Implementation of [Deep CORAL:
    Correlation Alignment for
    Deep Domain Adaptation](https://arxiv.org/abs/1607.01719)
    """

    def forward(self, x: torch.Tensor, y: torch.Tensor):
        """
        Arguments:
            x: features from one domain
            y: features from the other domain
        """
        embedding_size = x.shape[1]
        cx = covariance(x)
        cy = covariance(y)
        squared_fro_norm = torch.linalg.norm(cx - cy, ord="fro") ** 2
        return squared_fro_norm / (4 * (embedding_size ** 2))



import torch
from PIL import Image

from ..utils import common_functions as c_f


class BaseDataset(torch.utils.data.Dataset):
    def __init__(self, domain):
        super().__init__()
        self.domain = domain

    def __len__(self):
        return len(self.img_paths)

    def __getitem__(self, idx):
        label = self.labels[idx]
        img = Image.open(self.img_paths[idx]).convert("RGB")
        if self.transform is not None:
            img = self.transform(img)
        return img, label

    def __repr__(self):
        extra_repr = f"domain={self.domain}\nlen={str(self.__len__())}"
        return c_f.nice_repr(self, extra_repr, {"transform": self.transform})



import torch
import torch.nn as nn
import torch.nn.functional as F


# from https://arxiv.org/pdf/1409.7495.pdf
class MNISTFeatures(nn.Module):
    """
    A small convnet for extracting features
    from MNIST.
    """

    def __init__(self):
        """ """
        super().__init__()
        self.conv1 = nn.Conv2d(3, 32, 5, 1)
        self.conv2 = nn.Conv2d(32, 48, 5, 1)
        self.fc = nn.Identity()

    def forward(self, x):
        """ """
        x = self.conv1(x)
        x = F.relu(x)
        x = F.max_pool2d(x, kernel_size=2, stride=2)
        x = self.conv2(x)
        x = F.relu(x)
        x = F.max_pool2d(x, kernel_size=2, stride=2)
        x = torch.flatten(x, start_dim=1)
        return self.fc(x)



import os


def check_img_paths(img_dir, img_paths, domain):
    for x in img_paths:
        x = os.path.relpath(x, img_dir)
        if not x.startswith(domain):
            raise ValueError(
                f"img_paths contains a path {x} to an image outside of domain {domain}"
            )


def check_length(cls, correct_length):
    x = len(cls)
    if x != correct_length:
        raise ValueError(f"len(self)={x} but should be {correct_length}")



from typing import List, Union

import numpy as np
import torch
from pytorch_metric_learning.distances import LpDistance
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f
from . import utils as l_u


class MMDLoss(torch.nn.Module):
    """
    Implementation of

    - [Learning Transferable Features with Deep Adaptation Networks](https://arxiv.org/abs/1502.02791)

    - [Deep Transfer Learning with Joint Adaptation Networks](https://arxiv.org/abs/1605.06636).
    """

    def __init__(
        self, kernel_scales: Union[float, torch.Tensor] = 1, mmd_type: str = "linear"
    ):
        """
        Arguments:
            kernel_scales: The kernel bandwidth is scaled by this amount.
                If a tensor, then multiple kernel bandwidths are used.
            mmd_type: 'linear' or 'quadratic'. 'linear' uses the linear estimate of MK-MMD.
        """
        super().__init__()
        self.kernel_scales = kernel_scales
        self.dist_func = LpDistance(normalize_embeddings=False, p=2, power=2)
        self.mmd_type = mmd_type
        if mmd_type == "linear":
            self.mmd_func = l_u.get_mmd_linear
        elif mmd_type == "quadratic":
            self.mmd_func = l_u.get_mmd_quadratic
        else:
            raise ValueError("mmd_type must be either linear or quadratic")

    # input can be embeddings or list of embeddings
    def forward(
        self,
        x: Union[torch.Tensor, List[torch.Tensor]],
        y: Union[torch.Tensor, List[torch.Tensor]],
    ) -> torch.Tensor:
        """
        Arguments:
            x: features or a list of features from one domain.
            y: features or a list of features from the other domain.
        Returns:
            MMD if the inputs are tensors, and Joint MMD (JMMD) if the inputs are lists of tensors.
        """
        xx, yy, zz, scale = l_u.get_mmd_dist_mats(x, y, self.dist_func)
        if torch.is_tensor(self.kernel_scales):
            s = scale[0] if c_f.is_list_or_tuple(scale) else scale
            self.kernel_scales = pml_cf.to_device(self.kernel_scales, s, dtype=s.dtype)

        if c_f.is_list_or_tuple(scale):
            for i in range(len(scale)):
                scale[i] = scale[i] * self.kernel_scales
        else:
            scale = scale * self.kernel_scales

        return self.mmd_func(xx, yy, zz, scale)

    def extra_repr(self):
        """"""
        return c_f.extra_repr(self, ["mmd_type", "kernel_scales"])
